// SpatialSoften2
// Copyright 2004 by Seva Redkin <redkin@smartchat.net.au>
//
// It is slightly modified version of original SpatialSoften
// from Avisynth

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .

#include "avisynth.h"

#define min(a,b)	((a)<(b)?(a):(b))
#define max(a,b)	((a)>(b)?(a):(b))

static const int MAX_DIAMETER = 65;

class SpatialSoften2 : public GenericVideoFilter {
  const unsigned luma_threshold, chroma_threshold;
  const int diameter;

public:
  SpatialSoften2(PClip _child, int _radius, unsigned _luma_threshold,
                 unsigned _chroma_threshold, IScriptEnvironment* env);

  PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);

  static AVSValue __cdecl Create(AVSValue args, void*, IScriptEnvironment* env)
  {
    return new SpatialSoften2(args[0].AsClip(), args[1].AsInt(),
                              args[2].AsInt(), args[3].AsInt(), env);
  }
};

SpatialSoften2::SpatialSoften2(PClip _child, int _radius, unsigned _luma_threshold,
                               unsigned _chroma_threshold, IScriptEnvironment* env)
  : GenericVideoFilter(_child), diameter(_radius*2+1),
      luma_threshold(_luma_threshold), chroma_threshold(_chroma_threshold)
{
  if (!vi.IsYUY2())
    env->ThrowError("SpatialSoften2: requires YUY2 input");

  if (luma_threshold > 0xffU || chroma_threshold > 0xffU)
    env->ThrowError("SpatialSoften2: luma/chroma threshold greater that 255");

  if (diameter > MAX_DIAMETER)
    env->ThrowError("SpatialSoften2: radius is too big");

  if ((env->GetCPUFlags() & CPUF_INTEGER_SSE) == 0)
    env->ThrowError("SpatialSoften2: requires Athlon MMX extensions/Intel SSE");
}

extern "C" void calc2PelsBig(unsigned int src, int x, int diameter,
                    const unsigned char ** line,
                    unsigned int luma_threshold, unsigned int chroma_threshold,
                    unsigned int * dst);
extern "C" void calc2PelsSmall(unsigned int src, int x, int diameter,
                    const unsigned char ** line,
                    unsigned int luma_threshold, unsigned int chroma_threshold,
                    unsigned int * dst);

PVideoFrame SpatialSoften2::GetFrame(int n, IScriptEnvironment* env)
{
  PVideoFrame src = child->GetFrame(n, env);
  PVideoFrame dst = env->NewVideoFrame(vi);

  const unsigned char* srcp = src->GetReadPtr();
  unsigned char* dstp = dst->GetWritePtr();
  int src_pitch = src->GetPitch();
  int dst_pitch = dst->GetPitch();
  int row_size = src->GetRowSize();

  const unsigned char* line[MAX_DIAMETER];
  if (diameter <= 15) {
    for (int y=0; y<vi.height; ++y) {
      for (int h=0; h<diameter; ++h)
        line[h] = &srcp[src_pitch * min(max(y+h-(diameter>>1), 0), vi.height-1)];

      int x;
      int edge = (diameter+1) & -4;
      for (x=0; x<edge; ++x)
        dstp[y*dst_pitch + x] = srcp[y*src_pitch + x];

      for (; x < row_size - edge; x+=4)
        calc2PelsSmall(*(unsigned int *)&srcp[y*src_pitch + x],
                       x, diameter, line, luma_threshold, chroma_threshold,
                       (unsigned int *)&dstp[y*dst_pitch + x]);

      for (; x<row_size; ++x)
        dstp[y*dst_pitch + x] = srcp[y*src_pitch + x];
    }
  }
  else {
    for (int y=0; y<vi.height; ++y) {
      for (int h=0; h<diameter; ++h)
        line[h] = &srcp[src_pitch * min(max(y+h-(diameter>>1), 0), vi.height-1)];

      int x;
      int edge = (diameter+1) & -4;
      for (x=0; x<edge; ++x)
        dstp[y*dst_pitch + x] = srcp[y*src_pitch + x];

      for (; x < row_size - edge; x+=4)
        calc2PelsBig(*(unsigned int *)&srcp[y*src_pitch + x],
                       x, diameter, line, luma_threshold, chroma_threshold,
                       (unsigned int *)&dstp[y*dst_pitch + x]);

      for (; x<row_size; ++x)
        dstp[y*dst_pitch + x] = srcp[y*src_pitch + x];
    }
  }

  return dst;
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit(IScriptEnvironment* env)
{
  env->AddFunction("SpatialSoften2","c[radius]i[luma_treshold]i[chroma_treshold]i",
                   SpatialSoften2::Create, 0);
  return 0;
}
